﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Amib.Threading;
using System.Threading;

namespace ExchangeService
{
    /// <summary>
    /// Manages the External and DB providers and runs multiple threads
    /// to read from External source and write to DB and vice versa. Basically
    /// this class does everything you need to know. From a windows
    /// service you need to run this by calling .Start(). From a console
    /// app you can call .Start() and .Wait()
    /// </summary>
    public class ExchangeManager
    {
        private System.Func<IExternalProvider> ExternalProviderBuilder;
        private System.Func<IDBQueueProvider> DBQueueProviderBuilder;
        private ILogger Logger;
        private SmartThreadPool ThreadPool;
        private int ExternalReaderThreadsCount;
        private int DBPollerThreadCount;
        private TimeSpan DBPollInterval;
        private TimeSpan ExternalReadInterval;
        private TimeSpan ExternalReadTimeout;
        private TimeSpan StopTimeout;
        private bool Stopped = false;

        public ExchangeManager(
            System.Func<IExternalProvider> externalProviderBuilder, 
            System.Func<IDBQueueProvider> dbProviderBuilder,
            ILogger logger,
            int noOfExternalServiceThreads,
            int noOfDBPollerThreads,
            TimeSpan dbPollInterval,
            TimeSpan externalReadInterval,
            TimeSpan externalReadTimeout,
            TimeSpan stopTimeout)
        {
            this.ExternalProviderBuilder = externalProviderBuilder;
            this.DBQueueProviderBuilder = dbProviderBuilder;
            this.Logger = logger;

            this.DBPollerThreadCount = noOfDBPollerThreads;
            this.ExternalReaderThreadsCount = noOfExternalServiceThreads;
            this.DBPollInterval = dbPollInterval;
            this.ExternalReadTimeout = externalReadTimeout;
            this.ExternalReadInterval = externalReadInterval;
            this.StopTimeout = stopTimeout;
        }

        public void Start()
        {
            this.Logger.Info("Starting...");
            InitializeSmartThreadPool();
            StartExternalReaderThreads();
            StartDBReaderThreads();            
        }

        public bool IsRunning()
        {
            return this.ThreadPool.ActiveThreads > 0;
        }
        private void InitializeSmartThreadPool()
        {
            this.ThreadPool = new SmartThreadPool();
            this.ThreadPool.MaxThreads = this.ThreadPool.MinThreads = 
                this.DBPollerThreadCount + this.ExternalReaderThreadsCount;
            
        }

        /// <summary>
        /// Launch the threads that will poll DB at certain intervals
        /// and if any message found in Inbox, it will use the External Provider
        /// to send the message to the Queue.
        /// </summary>
        private void StartDBReaderThreads()
        {
            var me = this;
            for (int i = 0; i < this.DBPollerThreadCount; i++)
            {                
                this.ThreadPool.QueueWorkItem(() =>
                    {
                        me.Logger.Debug("Starting DB Reader Thread " + i + "...");
                        
                        var dbQueueProvider = me.DBQueueProviderBuilder();
                        var externalProvider = me.ExternalProviderBuilder();

                        try
                        {                            
                            me.Logger.Debug("Initializing DB Provider...");
                            dbQueueProvider.Initialize();
                            me.Logger.Debug("Initializing External Provider...");
                            externalProvider.Connect();

                            while (!me.Stopped)
                            {
                                try
                                {
                                    var inboxItems = dbQueueProvider.DequeueInbox();
                                    if (inboxItems != null)
                                    {
                                        foreach (InboxItem item in inboxItems)
                                        {
                                            me.Logger.Debug("Sending message to External " + item.ItemID + "...");
                                            me.Logger.LogInboxItem(item);
                                            try
                                            {
                                                externalProvider.Send(new ExternalMessage
                                                    {
                                                        CorrelationID = item.CorrelationID,
                                                        Data = item.TextData
                                                    });
                                            }
                                            catch (ThreadAbortException)
                                            {
                                                throw;
                                            }
                                            catch (Exception x)
                                            {
                                                if (!externalProvider.TryRecoverFromFault(x))
                                                    return;
                                            }
                                            me.Logger.Debug("Sent message to External " + item.CorrelationID);
                                        }
                                    }
                                }
                                catch (ThreadAbortException)
                                {
                                    throw;
                                }
                                catch (Exception x)
                                {
                                    me.Logger.Error(x);
                                    if (!dbQueueProvider.TryRecoverFromFault(x))
                                        return;
                                }
                                

                                Thread.Sleep(me.DBPollInterval);
                            }
                        }
                        catch (ThreadAbortException)
                        {
                            me.Logger.Info("Thread aborted.");
                        }
                        catch (Exception x)
                        {
                            me.Logger.Error(x);
                        }
                        finally
                        {
                            try
                            {
                                me.Logger.Debug("Disposing External Provider...");
                                externalProvider.Dispose();
                            }
                            catch (Exception x)
                            {
                                me.Logger.Error(x);
                            }
                            finally
                            {
                                try
                                {
                                    me.Logger.Debug("Disposing DB Provider...");
                                    dbQueueProvider.Dispose();
                                }
                                catch (Exception x)
                                {
                                    me.Logger.Error(x);
                                }
                            }
                        }

                        me.Logger.Info("Exiting DB Reader thread " + i);
                    });

                Thread.Sleep(this.DBPollInterval);
            }
        }

        /// <summary>
        /// Launch the threds for reading from external service.
        /// </summary>
        private void StartExternalReaderThreads()
        {
            var me = this;
            for (int i = 0; i < this.ExternalReaderThreadsCount; i++)
            {
                me.ThreadPool.QueueWorkItem(() =>
                {
                    me.Logger.Debug("Starting DB Reader Thread " + i + "...");

                    var dbQueueProvider = me.DBQueueProviderBuilder();
                    var externalQueueProvider = me.ExternalProviderBuilder();

                    try
                    {
                        me.Logger.Debug("Initializing DB Provider...");
                        dbQueueProvider.Initialize();
                        me.Logger.Debug("Initializing External Provider...");
                        externalQueueProvider.Connect();

                        while (!me.Stopped)
                        {
                            var message = default(ExternalMessage);
                            try
                            {
                                // This is a blocking call, until message arrives or we get
                                // some exception, execution does not proceed
                                message = externalQueueProvider.ReadAndWait(this.ExternalReadTimeout);
                            }
                            catch (ThreadAbortException)
                            {
                                throw;
                            }
                            catch (Exception x)
                            {
                                if (!externalQueueProvider.TryRecoverFromFault(x))
                                    return;
                            }

                            if (message != null)
                            {
                                me.Logger.Debug("Received message: " + message.CorrelationID);
                                me.Logger.LogMessage(message);
                                try
                                {
                                    var outboxItem = new OutboxItem
                                    {
                                        CorrelationID = message.CorrelationID,
                                        TextData = message.Data,
                                        ItemDateTime = DateTime.Now
                                    };

                                    me.Logger.Debug("Enqueue outbox item: " + outboxItem.CorrelationID);
                                    me.Logger.LogOutboxItem(outboxItem);
                                    dbQueueProvider.EnqueueOutbox(outboxItem);
                                }
                                catch (ThreadAbortException)
                                {
                                    throw;
                                }
                                catch (Exception x)
                                {
                                    me.Logger.Error(x);
                                    if (!dbQueueProvider.TryRecoverFromFault(x))
                                        return;
                                }

                                me.Logger.Debug("Queued message in Outbox: " + message.CorrelationID);                                
                            }
                            else
                            {
                                me.Logger.Debug("Got null message from External.");
                            }

                            Thread.Sleep(this.ExternalReadInterval);
                        }
                    }
                    catch (ThreadAbortException)
                    {
                        me.Logger.Info("Thread aborted.");
                    }
                    catch (Exception x)
                    {
                        me.Logger.Error(x);
                    }
                    finally
                    {
                        try
                        {
                            me.Logger.Debug("Disposing External Provider...");
                            externalQueueProvider.Dispose();
                        }
                        catch (Exception x)
                        {
                            me.Logger.Error(x);
                        }
                        finally
                        {
                            try
                            {
                                me.Logger.Debug("Disposing DB Provider...");
                                dbQueueProvider.Dispose();
                            }
                            catch (Exception x)
                            {
                                me.Logger.Error(x);
                            }
                        }
                    }

                    me.Logger.Info("Exiting External Reader thread " + i);
                });

                Thread.Sleep(this.ExternalReadInterval);
            }
        }

        public void Wait(TimeSpan timeout)
        {
            this.ThreadPool.WaitForIdle(timeout);
        }

        public void Stop()
        {
            this.Stopped = true;
            // Give 10 seconds to each thread to close gracefully
            Thread.Sleep(this.StopTimeout);
            try
            {
                this.ThreadPool.Shutdown(true, this.StopTimeout);
                this.ThreadPool.Dispose();
            }
            finally
            {
                
            }
        }        
    }
}
