﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace ExchangeService
{
    public interface IExternalProvider : IDisposable
    {
        /// <summary>
        /// Connect to the External provider. For MQ, establish the connection
        /// to the MQ here. For http, open http connection here.
        /// </summary>
        void Connect();
        /// <summary>
        /// Disconnect from external provider. Eg close connection to MQ or http.
        /// </summary>
        void Disconnect();
        /// <summary>
        /// Disconnect and Connect. You can either just call Disconnect and Connect
        /// or implement some custom code for handling a reconnection scenario.
        /// Reconnect is called when there's some problem with connection, say connection
        /// dropped abnormally.
        /// </summary>
        void Reconnect();

        /// <summary>
        /// A blocking read call to the external service. This function should not
        /// return unless some data is available. So, for MQ, you call Read() which
        /// blocks thread until a message arrives. For http service, you can keep 
        /// polling an http endpoint here. 
        /// If you can't implement a blocking call, then if there's nothing to return,
        /// return null.
        /// The blocking call must not block more than the given timeout.
        /// </summary>
        /// 
        /// <returns>Null if no data. Otherwise an ExternalMessage</returns>
        ExternalMessage ReadAndWait(TimeSpan timeout);

        /// <summary>
        /// Send a message to destination. Destination can be a MQ.
        /// </summary>
        /// <param name="message">Message to send</param>
        void Send(ExternalMessage message);

        /// <summary>
        /// Check if the exception is one of those exception that
        /// can be recovered from. For example, connection dropped. 
        /// Try Reconnect() within this function. If there's exception
        /// from Reconnect(), then return false to indicate permanent 
        /// failure. Return true if the connection was recovered. 
        /// </summary>
        /// <param name="x"></param>
        /// <returns>True if the exception is recoverable. False if permanent exception.</returns>
        bool TryRecoverFromFault(Exception x);
    }
}
