﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data.SqlClient;

namespace ExchangeService
{
    public class SqlDBQueueProvider : IDBQueueProvider, IDisposable
    {
        private SqlConnection Connection;
        private string ConnectionString;

        public SqlDBQueueProvider(string connectionString)
        {
            this.ConnectionString = connectionString;
        }
        
        public IEnumerable<InboxItem> DequeueInbox()
        {
            using (SqlCommand command = new SqlCommand("DequeueInbox", this.Connection))
            {
                command.CommandType = System.Data.CommandType.StoredProcedure;
                using (SqlDataReader reader = command.ExecuteReader())
                {
                    List<InboxItem> inboxItemList = new List<InboxItem>();
                    while (reader.Read())
                    {
                        inboxItemList.Add(new InboxItem
                        {
                            ItemID = (int)reader["ItemID"],
                            CorrelationID = (string)reader["CorrelationID"],
                            ItemDateTime = (DateTime)reader["ItemDateTime"],
                            TextData = (string)reader["TextData"]
                        });
                    }

                    return inboxItemList;
                }
            }
        }

        public void EnqueueInbox(InboxItem item)
        {
            using (SqlCommand command = new SqlCommand("EnqueueInbox", this.Connection))
            {
                command.CommandType = System.Data.CommandType.StoredProcedure;
                command.Parameters.Add("CorrelationID", System.Data.SqlDbType.NVarChar).Value = item.CorrelationID;
                command.Parameters.Add("TextData", System.Data.SqlDbType.NVarChar).Value = item.TextData;
                if (command.ExecuteNonQuery() <= 0)
                    throw new ApplicationException("EnqueueInbox did not insert one row");
            }
        }

        public void EnqueueOutbox(OutboxItem item)
        {
            using (SqlCommand command = new SqlCommand("EnqueueOutbox", this.Connection))
            {
                command.CommandType = System.Data.CommandType.StoredProcedure;
                command.Parameters.Add("CorrelationID", System.Data.SqlDbType.NVarChar).Value = item.CorrelationID;
                command.Parameters.Add("TextData", System.Data.SqlDbType.NVarChar).Value = item.TextData;
                if (command.ExecuteNonQuery() <= 0)
                    throw new ApplicationException("EnqueueOutbox did not insert one row");                              
            }
        }

        public OutboxItem RetrieveFromOutbox(string correlationId)
        {
            using (SqlCommand command = new SqlCommand("RetrieveFromOutbox", this.Connection))
            {
                command.CommandType = System.Data.CommandType.StoredProcedure;
                command.Parameters.Add("@CorrelationID", System.Data.SqlDbType.NVarChar).Value = correlationId;

                using (SqlDataReader reader = command.ExecuteReader())
                {
                    if (reader.Read())
                    {
                        return new OutboxItem
                        {
                            ItemID = (int)reader["ItemID"],
                            CorrelationID = (string)reader["CorrelationID"],
                            ItemDateTime = (DateTime)reader["ItemDateTime"],
                            TextData = (string)reader["TextData"]
                        };
                    }
                    else
                    {
                        return null;
                    }
                }
            }
        }

        public void Dispose()
        {
            this.Connection.Dispose();            
        }

        public void Initialize()
        {
            this.Connection = new SqlConnection(this.ConnectionString);
            this.Connection.Open();

        }


        public bool TryRecoverFromFault(Exception x)
        {
            if (x is SqlException)
            {
                SqlException sqlException = x as SqlException;

                if (sqlException.Errors.Count > 0)
                    return true;
                else
                {
                    try
                    {
                        Reinitialize();
                        return true;
                    }
                    catch
                    {
                        return false;
                    }
                }
            }
            else
            {
                try
                {
                    Reinitialize();
                    return true;
                }
                catch
                {
                    return false;
                }
            }
        }

        public void Reinitialize()
        {
            try
            {
                this.Connection.Close();
            }
            catch
            {
            }
            finally
            {
                this.Connection.Open();                
            }
        }
    }
}
