﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ExchangeService;
using Xunit;
using System.Threading;
using Moq;
using FluentAssertions;

namespace ExchangeServiceTests
{
    public class IntegrationTestExchangeServiceManager
    {
        private readonly static TimeSpan TestRunTime = TimeSpan.FromSeconds(10);

        [Fact]
        public void ShouldQueueMessagesReceivedFromExternalIntoInbox()
        {
            var messages = new List<ExternalMessage>();

            var manager = CreateManager(() =>
                {
                    var externalMock = new Mock<IExternalProvider>();
                    externalMock.Setup<ExternalMessage>(m => m.ReadAndWait(Config.ExternalReadTimeout))
                        .Returns(() =>
                        {
                            var message = new ExternalMessage
                            {
                                CorrelationID = Guid.NewGuid().ToString(),
                                Data = new string(Enumerable.Repeat<char>('x', new Random().Next(100 * 1024, 500 * 1024)).ToArray())
                            };
                            messages.Add(message);
                            return message;
                        });
                    return externalMock.Object;
                }, 
                () => new SqlDBQueueProvider(Config.ConnectionString));
            
            manager.Start();
            Thread.Sleep(TestRunTime);
            manager.Stop();


            messages.Count.Should().BePositive();

            // All the messages that has been returned from ReadAndWait callshould be 
            // in outbox by now.
            using (IDBQueueProvider dbProvider = new SqlDBQueueProvider(Config.ConnectionString))
            {
                dbProvider.Initialize();
                messages.ForEach(message =>
                    {
                        var outboxItem = dbProvider.RetrieveFromOutbox(message.CorrelationID);

                        outboxItem.Should().NotBeNull();
                        outboxItem.CorrelationID.Should().Be(message.CorrelationID);
                    });
            }
        }

        [Fact]
        public void ShouldSendEachMessageToExternalThatIsQueuedInInbox()
        {
            var inboxItemsQueued = new List<InboxItem>();
            var messagesSent = new List<ExternalMessage>();

            var manager = CreateManager(() =>
                {
                    var mock = new Mock<IExternalProvider>();
                    mock.Setup(m => m.Send(It.IsAny<ExternalMessage>()))
                        .Callback<ExternalMessage>(m => messagesSent.Add(m));
                    return mock.Object;   
                }, () => new SqlDBQueueProvider(Config.ConnectionString));

            manager.Start();

            // Store some messages in inbox that should be picked up and sent
            using (IDBQueueProvider dbProvider = new SqlDBQueueProvider(Config.ConnectionString))
            {
                dbProvider.Initialize();

                for (int i = 0; i < 10; i++)
                {
                    var inboxItem = new InboxItem
                    {
                        CorrelationID = Guid.NewGuid().ToString(),
                        TextData = new string(Enumerable.Repeat<char>('x', new Random().Next(100 * 1024, 500 * 1024)).ToArray())
                    };
                    inboxItemsQueued.Add(inboxItem);
                    dbProvider.EnqueueInbox(inboxItem);                    
                }
            }

            Thread.Sleep(TestRunTime);

            manager.Stop();

            inboxItemsQueued.ForEach(inboxItem =>
                {
                    messagesSent.Should().Contain(m => m.CorrelationID == inboxItem.CorrelationID &&
                        m.Data == inboxItem.TextData);
                });
        }


        private ExchangeManager CreateManager(Func<IExternalProvider> externalBuilder, 
            Func<IDBQueueProvider> dbBuilder)
        {
            return new ExchangeManager(externalBuilder,
                dbBuilder,
                new DummyLogger(),
                Config.ExternalThreadCount,
                Config.DBThreadCount,
                Config.DBPollInterval,
                Config.ExternalReadInterval,
                Config.ExternalReadTimeout,
                Config.StopTimeout);
        }
    }
}
