﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ExchangeService;
using Xunit;
using FluentAssertions;
using System.Threading;
using Moq;

namespace ExchangeServiceTests
{
    public class UnitTestExchangeServiceManager
    {
        private readonly static TimeSpan TestRunTime = TimeSpan.FromSeconds(10);
        [Fact]
        public void ShouldCreateConfiguredNumberOfProvidersAndDisposeThem()
        {
            int externalInitialized = 0, externalDisposed = 0, dbInitialized = 0, dbDisposed = 0;

            var manager = CreateServiceManager(() =>
            {
                var externalProviderMock = new Mock<IExternalProvider>();
                externalProviderMock.Setup(m => m.Connect()).Callback(() => externalInitialized++);
                externalProviderMock.Setup(m => m.Dispose()).Callback(() => externalDisposed++);
                return externalProviderMock.Object;
            }, () =>
            {
                var dbProviderMock = new Mock<IDBQueueProvider>();
                dbProviderMock.Setup(d => d.Initialize()).Callback(() => dbInitialized++);
                dbProviderMock.Setup(d => d.Dispose()).Callback(() => dbDisposed++);
                return dbProviderMock.Object;
            });

            manager.Start();
            Thread.Sleep(TestRunTime);
            manager.Stop();

            externalInitialized.Should().Be(Config.ExternalThreadCount * 2);
            externalDisposed.Should().Be(Config.ExternalThreadCount * 2);
            dbInitialized.Should().Be(Config.DBThreadCount * 2);
            dbDisposed.Should().Be(Config.DBThreadCount * 2);
        }

        [Fact]
        public void ShouldReadFromExternalAndPollFromDBFromAllThreads()
        {
            int externalRead = 0, dbPoll = 0;
            var mocks = new List<Mock>();
            var manager = CreateServiceManager(() =>
            {
                var externalProviderMock = new Mock<IExternalProvider>();
                externalProviderMock.Setup(m => m.ReadAndWait(Config.ExternalReadTimeout))
                    .Callback(() => 
                    {
                        Thread.Sleep(5000);
                        externalRead++;                        
                    });
                
                return externalProviderMock.Object;
            }, () =>
            {
                var dbProviderMock = new Mock<IDBQueueProvider>();
                dbProviderMock.Setup(d => d.DequeueInbox()).Callback(() => 
                    {
                        Thread.Sleep(5000);
                        dbPoll++;
                    });

                return dbProviderMock.Object;
            });

            manager.Start();
            Thread.Sleep(TestRunTime);
            manager.Stop();

            externalRead.Should().BeGreaterOrEqualTo(Config.ExternalThreadCount);
            dbPoll.Should().BeGreaterOrEqualTo(Config.DBThreadCount);            
        }
        
        [Fact]
        public void ShouldSendEachMessageToExternalThatIsQueuedInDB()
        {
            var dbDequeueCount = 0;
            var messagesQueued = 0;
            var externalThreadIds = new List<int>();
            var dbThreadIds = new List<int>();

            var message = new InboxItem
            {
                CorrelationID = Guid.NewGuid().ToString(),
                TextData = new string(Enumerable.Repeat<char>('X', 500000).ToArray())
            };
            var manager = CreateServiceManager(() =>
            {
                var externalProviderMock = new Mock<IExternalProvider>();
                externalProviderMock.Setup(mock => mock.Send(It.Is<ExternalMessage>(m => m.CorrelationID == message.CorrelationID
                    && m.Data == message.TextData)))
                    .Callback(() =>
                    {
                        var currentThreadId = Thread.CurrentThread.ManagedThreadId;
                        if (!externalThreadIds.Contains(currentThreadId))
                            externalThreadIds.Add(currentThreadId);

                        messagesQueued++;
                        Thread.Sleep(200);
                    })
                    .Verifiable();

                return externalProviderMock.Object;
            }, () =>
            {
                var dbProviderMock = new Mock<IDBQueueProvider>();
                dbProviderMock.Setup<IEnumerable<InboxItem>>(d => d.DequeueInbox())
                    .Returns(new InboxItem[] { message })
                    .Callback(() =>
                    {
                        var currentThreadId = Thread.CurrentThread.ManagedThreadId;
                        if (!dbThreadIds.Contains(currentThreadId))
                            dbThreadIds.Add(currentThreadId);
                        dbDequeueCount++;
                        Thread.Sleep(200);                        
                    })
                    .Verifiable();

                return dbProviderMock.Object;
            });

            manager.Start();
            Thread.Sleep(TestRunTime);
            manager.Stop();

            dbDequeueCount.Should().Be(messagesQueued);
            dbThreadIds.Should().BeEquivalentTo(externalThreadIds);
            dbDequeueCount.Should().BeLessThan((int)(TestRunTime.TotalMilliseconds / Config.DBPollInterval.Milliseconds) * Config.DBThreadCount);
        }

        [Fact]
        public void ShouldQueueEveryMessageInDBReceivedFromExternal()
        {
            var dbEnqueueCount = 0;
            var messagesReturned = 0;
            var externalThreadIds = new List<int>();
            var dbThreadIds = new List<int>();

            var externalMessage = new ExternalMessage
            {
                CorrelationID = Guid.NewGuid().ToString(),
                Data = new string(Enumerable.Repeat<char>('X', 500000).ToArray())
            };
            var manager = CreateServiceManager(() =>
            {
                var externalProviderMock = new Mock<IExternalProvider>();
                externalProviderMock.Setup<ExternalMessage>(m => m.ReadAndWait(Config.ExternalReadTimeout))
                    .Returns(externalMessage)
                    .Callback(() =>
                    {
                        var currentThreadId = Thread.CurrentThread.ManagedThreadId;
                        if (!externalThreadIds.Contains(currentThreadId))
                            externalThreadIds.Add(currentThreadId);

                        Thread.Sleep(TimeSpan.FromSeconds(Config.ExternalReadTimeout.TotalSeconds / 4));
                        messagesReturned++;
                    })
                    .Verifiable();

                return externalProviderMock.Object;
            }, () =>
            {
                var dbProviderMock = new Mock<IDBQueueProvider>();
                dbProviderMock.Setup(d => d.EnqueueOutbox(It.IsAny<OutboxItem>()))
                    .Callback(() =>
                    {
                        var currentThreadId = Thread.CurrentThread.ManagedThreadId;
                        if (!dbThreadIds.Contains(currentThreadId))
                            dbThreadIds.Add(currentThreadId);
                        Thread.Sleep(200);
                        dbEnqueueCount++;
                    })
                    .Verifiable();

                return dbProviderMock.Object;
            });

            manager.Start();
            Thread.Sleep(TestRunTime);
            manager.Stop();

            dbEnqueueCount.Should().Be(messagesReturned);
            dbThreadIds.Should().BeEquivalentTo(externalThreadIds);
            messagesReturned.Should().BeLessThan((int)(TestRunTime.TotalMilliseconds / Config.ExternalReadInterval.Milliseconds) * Config.DBThreadCount);
        }

        private static ExchangeManager CreateServiceManager(
            Func<IExternalProvider> mqProvider, 
            Func<IDBQueueProvider> dbProvider)
        {
            return new ExchangeManager(
                mqProvider ?? (() => new ExternalProviderStub()),
                dbProvider ?? (() => new SqlDBQueueProvider(Config.ConnectionString)),
                new DummyLogger(), 
                Config.ExternalThreadCount, 
                Config.DBThreadCount, 
                Config.DBPollInterval,
                Config.ExternalReadInterval,
                Config.ExternalReadTimeout,
                Config.StopTimeout);
        }
    }
}
